#!/usr/bin/python
#
# BIND IPV6 classes
#
# Jason Vas Dias <jvdias@redhat.com>, Red Hat Inc. Sep 2004

import DNS
import string
import re

REVERSE_ARPA=-2
REVERSE_INT=-3
REVERSE_ARPA_ZONE='IP6.ARPA.'
REVERSE_INT_ZONE='IP6.INT.'
GUESS=-1
NIBBLE_FORMAT=1
HEX_FORMAT=2
BITLABEL_FORMAT=3


class Address:
    data=''
    address=None
    prefix=None
    allow_prefix=False
    format=None
    errors=[]
        
    def bitlabel_in(self, data=None):
        if(data==None):
            data = self.data
        if (data == None) or (len(data)==0):
            return
        self.errors=[]
        last_address=self.address
        pfx_i=data.rfind('/')
        prefix=None
        self.address=None
        if pfx_i != -1:
            try:
                prefix=int(data[pfx_i+1:-1])
                self.prefix = prefix
            except ValueError:
                errors.append([-1,_('prefix value is not decimal')])
                return
            if (prefix < 0) or (prefix > 256):
                errors.append([-1,_('invalid prefix value')])            
        bits_in=0
        if data[2] not in 'box':
            quad = data[2:pfx_i]
            octs = quad.split('.')
            if len(octs) != 4:
                return
            self.address=''
            for o in octs:
                v=int(o)
                if( (v<0) or (v>255) ):
                    errors.append([-1,"invalid quad octet"])
                    address = last_address
                    return
                self.address += ("%.2x" % v)
            bits_in = 32
        else:
            radix=data[2]
            radii={'b':[2,4,1],'o':[8,4,3],'x':[16,1,4]}
            label=data[3:pfx_i]
            self.address=''
            for i in range(0,len(label),radii[radix][1]):
                d = label[i:i+radii[radix][1]]
                bits_in += len(d) * radii[radix][2] 
                for pd in range(0, radii[radix][1] - len(d)):
                    d+='0'
                try:
                    self.address += "%x" % int(d,radii[radix][0])
                except ValueError:
                    self.address = last_address
                    errors.append([-1,"Invalid bitlabel label"])
                    return
            digits_required = bits_in/4
            if digits_required % 4:
                digits_required += 1
            self.address = self.address[:digits_required]
        if (prefix != None):
            if(bits_in > prefix):
                digits_over = (bits_in - prefix)/4
                if( digits_over > 0):
                    self.address = self.address[:-digits_over]
            elif( bits_in < prefix ):
                digits_under =(prefix - bits_in)/4
                for i in range (0, digits_under):
                    self.address += '0'
        self.data=data
        
    def bitlabel_out(self):
        if( self.address == None):
            return None
        pfx = ''
        if( self.prefix != None ):
            pfx = '/' + str(self.prefix)            
        return r'\[x' + self.address + pfx + ']'        
        
    def hex_in(self, data=None):
        if(data==None):
            data = self.data
        if (data == None) or (len(data) == 0):
            return
        self.errors=[]
        last_address=self.address
        data_copy=data
        i = data.find('::')
        if i > -1:
            if(len(data) > (i+2)):
                if(data[i+2] == ':' ):
                    self.errors.append([-1,_("Invalid IPv6 address containing a sequence of more than two ':'s")])
                    return
                if data[i+2:].find('::') > -1:
                    self.errors.append([-1,_("Invalid IPv6 address containing more than one '::' sequence")])
                    return
        i = data.find('/')        
        prefix=None
        if (i > -1) and (len(data)>i):
            if not self.allow_prefix :
                self.errors.append([-1,_("Prefix address not allowed in this context")])
                return
            pfx=data[i+1:]
            try:
                prefix = int(pfx)
            except ValueError:
                self.errors.append([-1,_("Invalid non-decimal prefix: %s") % pfx])
                return
            data=data[:i]
        quad_re=re.compile(r':(([0-9]){1,3})((\.([0-9]){1,3}){3})$')
        m=quad_re.search(data)
        quad = ''
        hexs=8
        if( m != None ):
            quad = data[m.start()+1:]
            data = data[:-m.start()]
            hexs=6
        hex_re=re.compile(r'^([0-9a-fA-F]{1,4})?(:([0-9a-fA-F]{0,4})){0,'+str(hexs)+'}$')
        if not hex_re.match(data):
            self.errors.append([-1,_("Invalid IPv6 address %s") % str(hexs)])
            return
        self.address=''
        if data == '::':
            data=''
        else:
            if data[0:2]=='::':
                data=data[1:]
            if data[-2:]=='::':
                data=data[:-1]
        hexes=data.split(":")
        n_hexes=len(hexes)
        for hex in hexes:
            n_zero_words=1
            if len(hex)==0:
                n_zero_words += hexs - n_hexes
            for d in range(0,(n_zero_words * 4)-len(hex)):
                hex='0'+hex
            self.address += hex
        if len(quad):
            for oct in quad.split('.'):
                self.address += '%0.2x' % int(oct)
        self.prefix = prefix
        if len(self.address) != 32:
            self.errors.append([-1,_('Invalid length for IPv6 address')])
            self.address = None
            self.prefix = None
        self.data=data_copy
        
    def hex_out(self, data=None):
        if( self.address==None ) or (len(self.address)==0):
            return
        debug('hex_out: ', self.address)
        zseq=[]
        s=None
        i=0
        for c in self.address:
            if c == '0':
                if(s == None):
                    s=i
            elif s != None:
                zseq.append([s,i])
                s=None
            i+=1
        if( (s != None) and (i>s) ):
            zseq.append([s,i])
        if(len(zseq)==0) and ( s != None ) and (i == len(self.address)):
            zseq=[[s,len(self.address)]]
        debug('hex_out: initial zseq:', zseq)
        i=0
        l=0        
        longest=None
        for s in zseq:
            if (s[1] - s[0]) > l:
                l=s[1] - s[0]
                if l >= 4:
                    longest=i
            i+=1
        start=None
        end=None
        debug('hex_out: longest:',longest)
        if longest != None:
            s = zseq[longest]
            start = s[0] / 4
            end = s[1] / 4
            if (s[0] % 4) > 0:
                start += 1
            start *= 4
            end *= 4
            if (end - start) < 4:
                start=None
                end=None
        begin = False
        output=''
        debug('hex_out: start:', start, ' end:', end)
        for i in range(0,len(self.address),4):
            if (not begin) and (start != None) and (start <= i):
                begin = True
                if len(output) and (output[-1] == ":"):
                    output += ':'
                else:
                    output += '::'
            elif( ((not begin) and (i <= start)) or ((end != None) and (i >= end)) ) :
                word=self.address[i:i+4]
                out=''
                for c in word:
                    if (c != '0') or len(out):
                        out+=c
                if (len(out)):
                    output += out + ':'
                else:
                    output += word + ':'
        debug('hex_out: output 1:', output)
        if(output[-1] == ':'):
            output= output[:-1]
        i = output.rfind(':')
        if ( i > -1 ):            
            trailingZeros=True
            debug('hex_out: trailing zeros? : ',output[i+1:])
            for c in output[i+1:][:]:
                if( c != '0' ):
                    trailingZeros=False
            if trailingZeros:
                output = output[0:i+1]
                debug('hex_out: trailing zeros :', output)
        if( self.prefix != None ):
            output += '/' + str(self.prefix)
        if( output == '::' ) and (len(self.address) < 32):
            output += '/' + str(int( len(self.address) * 4 ))
        return output

    def split_hex(self, addr=None):
        a=''
        if( addr == None):
            a = self.address
        else:
            a = addr
        if( len(a) < 32 ):
            for i in range(len(a), 32):
                a += '0'
        hexs=[]
        for i in range(0,32,4):
            hexs.append(a[i:i+4])
        return hexs
    
    def nibble_in(self, data=None):
        if(data==None):
            data=self.data
        if( data == None) or (len(data)==0):
            self.errors.append([-1,'Null IPV6 nibble-format address'])
            return
        i = data.find('/')
        if i > -1:
            if not self.allow_prefix :
                self.errors.append([-1,_("Prefix address not allowed in this context")])
                return
            try:
                self.prefix = int(data[i+1:])
            except ValueError:
                self.errors.append([-1,'non-decimal prefix value'])
                return
            data=data[:i]
        nre=re.compile('^([0-9A-Fa-f])(\.[0-9A-Fa-f]){0,31}$')
        if not nre.match(data):
            self.errors.append([-1,'Invalid IPV6 nibble-format address'])
            return
        self.address=string.join(data.split('.'),'')
        if ( self.prefix != None ):
            digits_required = self.prefix / 4
            if  self.prefix % 4:
                digits_required += 1
            if( len(self.address) < digits_required ):
                for i in range (0, digits_required - len(self.address)):
                    self.address += '0'
            else:
                self.address = self.address[:digits_required]

    def nibble_out(self):
        output=''
        if (self.address == None) or (len(self.address) == 0):
            return output
        for c in self.address:
            output += c + '.'
        if(len(output)):
            output=output[:-1]
        return output

    def coalesce(self,data):
        if (data==None) or (len(data)<3):
            return None
        if (data.find(r'.\[') > -1) or ( data.find(r'].') >-1 ):
            nre=re.compile('^([0-9A-Fa-f])(\.[0-9A-Fa-f]){0,31}$')
            full_address=''
            address=data
            blb=data.find(r'\[')
            while blb > -1:
                pre=address[:blb]
                if len(pre):
                    if(pre[-1] == '.'):
                        pre = pre[:-1]
                    if len(pre):
                        if not nre.match(pre):
                            return None
                        hpre = string.join(pre.split('.'),'')
                        if( (len(full_address) + len(hpre))  < 32 ):
                            full_address+=hpre
                        else:
                            return None
                ble = address[blb:].find(']')
                if( ble < 2 ):
                    debug( 'ble < 2')
                    return None
                bitlabel = address[blb:blb+ble+1]
                a = Address(bitlabel)
                if a.address == None:
                    self.errors += a.errors
                    return None
                if ( len(a.address) + len(full_address) ) > 32:
                    return None
                full_address += a.address
                address = address[blb+ble+1:]
                if( len(address) == 0):
                    self.address = full_address
                    self.format = NIBBLE_FORMAT
                    return True
                if address[0]=='.':
                    address=address[1:]
                blb = address.find(r'\[')
                if ( blb==-1 ):
                    if len(address) and not nre.match(address):
                        return None
                    ha=string.join(address.split('.'),'')
                    if( (len(full_address) + len(ha))  <= 32 ):
                        full_address+=ha
                    else:
                        return None
            self.address = full_address
            self.format=NIBBLE_FORMAT
            return True
        return None
                                
    def input(self, data, format=GUESS, allow_prefix = False):
        self.allow_prefix = allow_prefix
        self.address=None
        self.prefix=None
        self.errors=[]
        if (data==None) or (len(data) == 0):
            return
        self.data = data        
        if( self.coalesce(data) != None ):            
            if self.address != None:
                return True
            return False                
        if (format==BITLABEL_FORMAT) or ( (data[0:2] == r'\[') and (data[-1] == ']') ):
            self.bitlabel_in()
            if( self.address != None):
                self.format = BITLABEL_FORMAT
        elif (format == HEX_FORMAT) or ( data.find(':') > -1 ):
            self.hex_in()
            if( self.address != None):
                self.format = HEX_FORMAT
        else:
            if (format==NIBBLE_FORMAT) or ((data.find('.') > -1) or (len(data) == 1)):
                self.nibble_in()
                if( self.address != None):
                    self.format = NIBBLE_FORMAT
            else:
                self.hex_in()
                if( self.address != None):
                    self.format = HEX_FORMAT
        if( self.address != None):
            self.errors = None
        return self.address != None
        
    def __init__(self, data, format=GUESS, allow_prefix = False):
        self.input(data,format,allow_prefix)

    def __str__(self):
        if(( self.address != None )  and (self.format != None)):
            if(self.format == HEX_FORMAT):
                return self.hex_out()
            elif (self.format == NIBBLE_FORMAT):
                return self.nibble_out()
            elif (self.format == BITLABEL_FORMAT):
                return self.bitlabel_out()
        else:
            return ''

    def int(s):
        if(s.address != None):
            return long(s.address,16)
        return None

class AAAA(DNS.Data):
    address = None
    def __init__(self, addr):
        if (addr==None) or (len(addr) == 0):
            addr='::'
        self.address = Address(addr, HEX_FORMAT)
    
    def __str__(self):
        return str(self.address)
    

class A6(DNS.Data):
    def __init__(self,zone,tokens):
        debug( 'IPV6.A6.__init__', zone, tokens)
        self.prefix = 0
        self.address= DNS.IPV6.Address('0',NIBBLE_FORMAT)
        self.prefix_name=DNS.Name('',zone)
        self.tokens=tokens
        self.zone=zone
        if len(tokens) < 2:            
            return
        try:
            self.prefix = int(tokens[0])
        except ValueError:
            self.prefix=0
            return
        self.address = Address(tokens[1],HEX_FORMAT)
        if self.address.address == None:
            if( self.prefix == 128 ):
                self.prefix_name = DNS.Name(tokens[1],zone)
            return
        if( self.prefix > 0 ):
            self.prefix_name = DNS.Name(tokens[2],zone)
        else:
            self.prefix_name = None
        
    def out(self):
        s=' ' + str(self.prefix)
        if( (self.prefix < 128 ) and (self.address != None) and (self.address.address != None) ):
            s += ' ' + str(self.address)
        if( (self.prefix > 0) and (self.prefix_name != None) and (len(self.prefix_name.data))):
            s += ' ' + self.prefix_name.relative_label()
        return s
        
    def __str__(self):
        return self.out()
    
    def label(self):
        n=''
        pfx=''
        if self.prefix > 0:
            pfx='/'+str(self.prefix)
            n=pfx
            if (self.prefix_name != None) and (len(self.prefix_name.data)>0):
                n='('+self.prefix_name.relative_label()+pfx+')'
        a='::'
        if (self.address != None) and (self.address.address != None) and (len(self.address.address)):
            a=self.address.hex_out()
        return n + a
    
def reverse(data):
    out=''
    p=''
    bl=''
    for i in range(len(data)-1,-1,-1):
        c = data[i]
        if c == ']':
            bl=c
            continue
        elif (p == '[') and (c == '\\' ):
            bl = c + bl
            out += bl
            bl=''
            continue
        if( len(bl) == 0):
            out += c
        else:
            bl = c + bl
        p=c
    return out

class ReverseARPAddress(DNS.Data):    
    addr=''
    address=None
    origin_address=None
    origin=None
    zone_origin=None

    def remove_arpa(self,n):
        if n[-len(REVERSE_ARPA_ZONE):].upper() == REVERSE_ARPA_ZONE:
            n = n[:-len(REVERSE_ARPA_ZONE)]
            if n[-1] == '.':
                n = n[:-1]        
        return n
    
    def __init__(self, addr, zone):
        self.addr=addr
        self.origin=zone.origin
        self.zone_origin=zone.zone_origin
        orig = self.remove_arpa(self.zone_origin)
        name = self.remove_arpa(self.name())
        self.origin_address = Address( reverse(orig) )
        self.address = Address( reverse(name) )        
        
    def name(self):
        if (self.addr == '@') or (self.addr == None) or (len(self.addr)==0):
            return self.origin    
        elif self.addr[-1] == '.':
            return self.addr
        elif self.origin != '.':
            return self.addr + '.' + self.origin
        elif(self.origin != None) and (len(self.origin)):
            return self.addr + self.origin
        else:
            return self.addr

    def value(self):
        return self.addr
    
    def __str__(self):
        return self.name()

    def label(self):
        return self.address.hex_out()

    def relative_label(self):
        a = self.origin_address.nibble_out()
        b = self.address.nibble_out()
        if(len(a) <= 0):
            debug( 'IPV6 LABEL - NO ORIGIN:',self.origin, self.zone_origin)
            return self.address.hex_out()
        rel=b[len(a)+1:]
        if( len(rel) == 0 ):
            return ''
        if( rel[0]=='.'):
            rel = rel[1:]
        if( rel[-1]=='.'):
            rel = rel[:-1]
        c = Address(rel,NIBBLE_FORMAT)
        if c.address == None:
            debug( 'IPV6 LABEL ERROR:', a, b, rel,c.errors)
        while(len(c.address)<32):
            c.address = '0' + c.address
#        print 'IPV6 LABEL:', '\n\tORIGIN:\t', a, '\n\tADDR\t:',b,'\n\tREL:\t',rel,'\n\t',c.address,c.hex_out()
        hex= c.hex_out()
        if(hex[:2] == '::'):
            hex=hex[2:]
        return hex

    def contains(s,ip):
        a=s.name()        
        return ((len(a)>0) and (len(ip)>0) and ( len(ip) > len(a) ) and ( ip[-(len(a)):].upper() == a.upper() ))

    def split_domain(s,n=None):
        if(n==None):
            n=reverse(s.address.nibble_out())
        a=n.split('.')
        hn=''
        dn=''
        debug( 'IPV6 SPLIT DOMAIN:',len(a))
        if( len(a) >= 24 ):
            hn=string.join(a[:16],'.')
            if( hn[-1]=='.'):
                hn=hn[:-1]
            dn=string.join(a[16:],'.')
            if dn[0]=='.':
                dn=dn[1:]
            if dn[-1] != '.':
                dn = dn +'.'
        elif( len(a) >=  16 ):
            hn=string.join(a[:8],'.')
            if( hn[-1]=='.'):
                hn=hn[:-1]
            dn=string.join(a[8:],'.')
            if dn[0]=='.':
                dn=dn[1:]
            if dn[-1] != '.':
                dn = dn +'.'
        elif( len(a) >= 8 ):
            hn=string.join(a[:4],'.')
            if( hn[-1]=='.'):
                hn=hn[:-1]
            dn=string.join(a[4:],'.')
            if dn[0]=='.':
                dn=dn[1:]
            if dn[-1] != '.':
                dn = dn +'.'          
        else:
            hn,dn=(a[0],string.join(a[1:],'.'))
            if(dn[0]=='.'):
                dn=dn[1:]
            if(dn[-1] != '.'):
                dn += '.'                
        debug( 'IPV6 SPLIT DOMAIN:','HN:',hn,'DN:',dn + REVERSE_ARPA_ZONE        )
        return (hn,dn + REVERSE_ARPA_ZONE)

    def relative_value(self,origin=None):
        if origin == None:
            if has_attr(self,'zone_origin'):
                origin = self.zone_origin
            else:
                origin = ''
        s = self.__str__()
        if (len(origin)>0) and (len(s) >= len(origin)) and (s[-len(origin):]==origin):
            s=s[:-len(origin)]
            if( len(s) == 0):
                s = '@'
            if(len(s) > 1) and (s[-1] == '.'):
                s=s[:-1]
        return s        

class ReverseINTAddress(DNS.Data):    
    addr=''
    address=None
    origin_address=None
    origin=None
    zone_origin=None

    def remove_arpa(self,n):
        if n[-len(REVERSE_INT_ZONE):].upper() == REVERSE_INT_ZONE:
            n = n[:-len(REVERSE_INT_ZONE)]
            if n[-1] == '.':
                n = n[:-1]        
        return n
    
    def __init__(self, addr, zone):
        self.addr=addr
        self.origin=zone.origin
        self.zone_origin=zone.zone_origin
        orig = self.remove_arpa(self.zone_origin)
        name = self.remove_arpa(self.name())
        self.origin_address = Address( reverse(orig) )
        self.address = Address( reverse(name) )        
        
    def name(self):
        if self.addr == '@':
            return self.origin    
        elif self.addr[-1] == '.':
            return self.addr
        elif self.origin != '.':
            return self.addr + '.' + self.origin
        else:
            return self.addr + self.origin

    def value(self):
        return self.relative_label()
    
    def __str__(self):
        return self.name()

    def label(self):
        return self.address.hex_out()

    def relative_label(self):
        a = self.origin_address.nibble_out()
        b = self.address.nibble_out()
        return b[len(a)+1:]            

    def split_domain(s,n=None):
        if(n==None):
            n=str(s)
        a=n.split('.')
        return (a[0],DNS.make_absolute(string.join(a[1:],'.')))

    def relative_value(self,origin=None):
        if origin == None:
            if has_attr(self,'zone_origin'):
                origin = self.zone_origin
            else:
                origin = ''
        s = self.__str__()
        if (len(origin)>0) and (len(s) >= len(origin)) and (s[-len(origin):]==origin):
            s=s[:-len(origin)]
            if( len(s) == 0):
                s = '@'
            if(len(s) > 1) and (s[-1] == '.'):
                s=s[:-1]
        return s        
